/*******************************************************************************
 * Copyright (c) 2009 TopQuadrant, Inc.
 * All rights reserved. 
 *******************************************************************************/
package org.topbraid.spin.inference;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.topbraid.spin.system.ARQFactory;
import org.topbraid.spin.system.SPINLabels;
import org.topbraid.spin.util.JenaUtil;
import org.topbraid.spin.util.ProgressMonitor;
import org.topbraid.spin.util.SPINUtil;
import org.topbraid.spin.vocabulary.SPIN;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.GraphListener;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.ontology.OntModel;
import com.hp.hpl.jena.ontology.OntModelSpec;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.shared.ReificationStyle;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;

/**
 * Static methods to find and execute spin:constructors for a given
 * set of Resources.
 * 
 * @author Holger Knublauch
 */
public class SPINConstructors {
	
	/**
	 * Runs the constructors on a List of Resources.
	 * @param queryModel  the model to query over
	 * @param instances  the instances to run the constructors of
	 * @param targetModel  the model that shall receive the new triples
	 * @param monitor  an optional progress monitor
	 */
	public static void construct(Model queryModel, List<Resource> instances, Model targetModel, ProgressMonitor monitor) {
		construct(queryModel, instances, targetModel, new HashSet<Resource>(), monitor);
	}

	
	/**
	 * Runs the constructors on a List of Resources.
	 * @param queryModel  the model to query over
	 * @param instances  the instances to run the constructors of
	 * @param targetModel  the model that shall receive the new triples
	 * @param reached  the Set of already reached Resources
	 * @param monitor  an optional progress monitor
	 */
	public static void construct(Model queryModel, List<Resource> instances, Model targetModel, Set<Resource> reached, ProgressMonitor monitor) {
		construct(queryModel, instances, targetModel, reached, null, monitor);
	}
	
	
	/**
	 * Runs the constructors on a List of Resources.
	 * @param queryModel  the model to query over
	 * @param instances  the instances to run the constructors of
	 * @param targetModel  the model that shall receive the new triples
	 * @param reached  the Set of already reached Resources 
	 * @param explanations  an (optional) explanations object
	 * @param monitor  an optional progress monitor
	 */
	public static void construct(Model queryModel, List<Resource> instances, Model targetModel, Set<Resource> reached, SPINExplanations explanations, ProgressMonitor monitor) {
		if(!instances.isEmpty()) {
			List<Resource> newResources = new ArrayList<Resource>();
			for(Resource instance : instances) {
				if(!reached.contains(instance)) {
					reached.add(instance);
					constructInstance(queryModel, instance, targetModel, newResources, explanations, monitor);
				}
			}
			construct(queryModel, newResources, targetModel, reached, explanations, monitor);
		}
	}
	

	/**
	 * Runs constructors for a single instance.
	 * @param queryModel  the model to query
	 * @param instance  the instance to run the constructors of
	 * @param targetModel  the model that will receive the new triples
	 * @param newResources  will hold the newly constructed instances
	 * @param monitor  an optional progress monitor
	 */
	public static void constructInstance(Model queryModel, Resource instance, Model targetModel, List<Resource> newResources, SPINExplanations explanations, ProgressMonitor monitor) {
		for(Statement s : JenaUtil.getStatementsList(instance.listProperties(RDF.type))) {
			Resource type = s.getResource();
			constructInstance(queryModel, instance, type, targetModel, newResources, new HashSet<Resource>(), explanations, monitor);
		}
	}

	
	/**
	 * Runs all constructors defined for a given type on a given instance.
	 * @param queryModel  the model to query
	 * @param instance  the instance to run the constructors of
	 * @param type  the class to run the constructors of
	 * @param targetModel  the model that will receive the new triples
	 * @param newResources  will hold the newly constructed instances
	 * @param reachedTypes  contains the already reached types
	 * @param monitor  an optional progress monitor
	 */
	public static void constructInstance(Model queryModel, Resource instance, Resource type, Model targetModel, List<Resource> newResources, Set<Resource> reachedTypes, ProgressMonitor monitor) {
		constructInstance(queryModel, instance, type, targetModel, newResources, reachedTypes, monitor);
	}
	

	/**
	 * Runs all constructors defined for a given type on a given instance.
	 * @param queryModel  the model to query
	 * @param instance  the instance to run the constructors of
	 * @param type  the class to run the constructors of
	 * @param targetModel  the model that will receive the new triples
	 * @param newResources  will hold the newly constructed instances
	 * @param reachedTypes  contains the already reached types
	 * @param explanations  the explanations (optional)
	 * @param monitor  an optional progress monitor
	 */
	public static void constructInstance(Model queryModel, Resource instance, Resource type, 
			Model targetModel, List<Resource> newResources, Set<Resource> reachedTypes, 
			SPINExplanations explanations, ProgressMonitor monitor) {
		
		// Run superclass constructors first
		for(Statement s : JenaUtil.getStatementsList(type.listProperties(RDFS.subClassOf))) {
			Resource superClass = s.getResource();
			if(!reachedTypes.contains(superClass)) {
				reachedTypes.add(superClass);
				constructInstance(queryModel, instance, superClass, targetModel, newResources, reachedTypes, explanations, monitor);
			}
		}
		
		List<String> queryStrings = new ArrayList<String>(SPINUtil.getQueryStrings(type, SPIN.constructor));
		Collections.sort(queryStrings);
		
		for(String queryString : queryStrings) {
			
			Query arqQuery = ARQFactory.get().createQuery(queryModel, queryString);
			
			if(arqQuery.isConstructType()) {
				
				QueryExecution qexec = ARQFactory.get().createQueryExecution(arqQuery, queryModel);
				SPINUtil.bindThis(qexec, instance);
				
				// Execute construct and remember the order in which triples were inserted
				// Note that this does not work yet since Jena appears to have random order
				Model resultModel = ModelFactory.createDefaultModel(ReificationStyle.Minimal);
				final List<Triple> triples = new LinkedList<Triple>();
				resultModel.getGraph().getEventManager().register(new GraphListener() {

					public void notifyAddTriple(Graph g, Triple t) {
						triples.add(t);
					}

					
					public void notifyDeleteTriple(Graph g, Triple t) {
					}

					
					public void notifyAddArray(Graph g, Triple[] triples) {
						for (int i = 0; i < triples.length; i++) {
							notifyAddTriple(g, triples[i]);
						}
					}


					public void notifyAddGraph(Graph g, Graph added) {
					}


					public void notifyAddIterator(Graph g, Iterator<Triple> it) {
						while (it.hasNext()) {
							Triple t = it.next();
							notifyAddTriple(g, t);
						}
					}


					public void notifyAddList(Graph g, List<Triple> triples) {
						notifyAddIterator(g, triples.iterator());
					}


					public void notifyDeleteArray(Graph g, Triple[] triples) {
						for (int i = 0; i < triples.length; i++) {
							notifyDeleteTriple(g, triples[i]);
						}
					}


					public void notifyDeleteGraph(Graph g, Graph removed) {
					}


					public void notifyDeleteIterator(Graph g, Iterator<Triple> it) {
						while (it.hasNext()) {
							Triple triple = it.next();
							notifyDeleteTriple(g, triple);
						}
					}


					public void notifyDeleteList(Graph g, List<Triple> list) {
						notifyDeleteIterator(g, list.iterator());
					}


					public void notifyEvent(Graph source, Object value) {
					}
				});
				qexec.execConstruct(resultModel);
				
				StringBuffer sb = new StringBuffer();
				sb.append("Inferred by SPIN constructor at class ");
				sb.append(SPINLabels.get().getLabel(type));
				sb.append(":\n\n" + queryString);
				String explanationText = sb.toString();

				// Add all new triples and any new resources
				for(Triple triple : triples) {
					Statement rs = queryModel.asStatement(triple);
					if(!targetModel.contains(rs)) {
						targetModel.add(rs);
						if(RDF.type.equals(rs.getPredicate())) {
							Resource subject = rs.getSubject();
							if(!newResources.contains(subject)) {
								newResources.add(subject);
							}
						}
						if(explanations != null) {
							explanations.put(triple, explanationText);
						}
					}
				}
			}
		}
	}
	
	
	/**
	 * Runs all constructors on all instances in a given model.
	 * @param queryModel  the query model
	 * @param targetModel  the model to write the new triples to
	 * @param monitor  an optional progress monitor
	 */
	public static void constructAll(Model queryModel, Model targetModel, ProgressMonitor monitor) {
		Set<Resource> classes = getClassesWithConstructor(queryModel);
		List<Resource> instances = new ArrayList<Resource>(getInstances(classes));
		OntModel ontModel = ModelFactory.createOntologyModel(OntModelSpec.OWL_MEM, queryModel);
		if(targetModel != queryModel) {
			ontModel.addSubModel(targetModel);
		}
		construct(ontModel, instances, targetModel, new HashSet<Resource>(), monitor);
	}
	

	/**
	 * Finds all classes that directly have a spin:constructor attached
	 * to it.
	 * @param model  the Model to operate on
	 * @return a Set of classes
	 */
	public static Set<Resource> getClassesWithConstructor(Model model) {
		Set<Resource> results = new HashSet<Resource>();
		for(Property property : getConstructorProperties(model)) {
			StmtIterator it = model.listStatements(null, property, (RDFNode)null);
			while(it.hasNext()) {
				results.add(it.next().getSubject());
			}
		}
		return results;
	}
	
	
	private static Iterable<Property> getConstructorProperties(Model model) {
		List<Property> results = new ArrayList<Property>();
		for(Resource r : JenaUtil.getAllSubProperties(model.getProperty(SPIN.constructor.getURI()))) {
			results.add(model.getProperty(r.getURI()));
		}
		results.add(model.getProperty(SPIN.constructor.getURI()));
		return results;
	}
	
	
	private static Set<Resource> getInstances(Collection<Resource> classes) {
		Set<Resource> results = new HashSet<Resource>();
		for(Resource cls : classes) {
			results.addAll(JenaUtil.getAllInstances(cls));
		}
		return results;
	}
	
	
	/**
	 * Checks whether a given class or a superclass thereof has a
	 * constructor.
	 * @param cls  the class to check
	 * @return true if cls has a constructor
	 */
	public static boolean hasConstructor(Resource cls) {
		for(Property property : getConstructorProperties(cls.getModel())) {
			if(cls.hasProperty(property)) {
				return true;
			}
			else {
				for(Resource superClass : JenaUtil.getAllSuperClasses(cls)) {
					if(superClass.hasProperty(property)) {
						return true;
					}
				}
			}
		}
		return false;
	}
}
