/*******************************************************************************
 * Copyright (c) 2009 TopQuadrant, Inc.
 * All rights reserved. 
 *******************************************************************************/
package org.topbraid.spin.model.print;

import java.util.List;
import java.util.Map;

import org.topbraid.spin.system.ExtraPrefixes;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.compose.MultiUnion;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.impl.Util;
import com.hp.hpl.jena.sparql.util.NodeToLabelMap;


/**
 * A simple implementation of PrintContext that operates on a StringBuilder.
 * 
 * By default this is using prefixes but not extra prefixes.
 * 
 * @author Holger Knublauch
 */
public class StringPrintContext implements PrintContext {
	
	private int indentation;
	
	protected String indentationString = "    ";
	
	private NodeToLabelMap nodeToLabelMap;
	
	private boolean nested;
	
	private boolean printPrefixes;
	
	private StringBuilder sb;
	
	private boolean useExtraPrefixes;
	
	private boolean usePrefixes = true;
	
	
	public StringPrintContext() {
		this(new StringBuilder());
	}
	
	
	public StringPrintContext(StringBuilder sb) {
		this.sb = sb;
	}
	
	
	public PrintContext clone() {
		StringPrintContext cl = new StringPrintContext(sb);
		cl.setIndentation(getIndentation());
		cl.setNested(isNested());
		cl.setUseExtraPrefixes(getUseExtraPrefixes());
		cl.setUsePrefixes(getUsePrefixes());
		cl.sb = this.sb;
		return cl;
	}
	

	public int getIndentation() {
		return indentation;
	}
	
	
	public NodeToLabelMap getNodeToLabelMap() {
		if(nodeToLabelMap == null) {
			nodeToLabelMap = new NodeToLabelMap();
		}
		return nodeToLabelMap;
	}


	public boolean getPrintPrefixes() {
		return printPrefixes;
	}


	public String getString() {
		return sb.toString();
	}
	
	
	public StringBuilder getStringBuilder() {
		return sb;
	}


	public boolean getUseExtraPrefixes() {
		return useExtraPrefixes;
	}
	
	
	public boolean getUsePrefixes() {
		return usePrefixes;
	}
	
	
	public boolean isNested() {
		return nested;
	}
	
	
	public void print(String str) {
		sb.append(str);
	}


	public void printIndentation(int depth) {
		for(int i = 0; i < depth; i++) {
			print(indentationString);
		}
	}


	public void printKeyword(String str) {
		print(str);
	}


	public void println() {
		print("\n");
	}


	public void printVariable(String str) {
		print("?" + str);
	}


	public void printURIResource(Resource resource) {
		if(getUsePrefixes()) {
			String qname = qnameFor(resource);
			if(qname != null) {
				print(qname);
				return;
			}
			else if(getUseExtraPrefixes()) {
				Map<String,String> extras = ExtraPrefixes.getExtraPrefixes();
				for(String prefix : extras.keySet()) {
					String ns = extras.get(prefix);
					if(resource.getURI().startsWith(ns)) {
						print(prefix);
						print(":");
						print(resource.getURI().substring(ns.length()));
						return;
					}
				}
			}
		}
		print("<");
		print(resource.getURI());
		print(">");
	}
	

	/**
	 * Work-around for a bug in Jena: Jena would use the default
	 * namespace of an imported Graph in a MultiUnion.
	 * @param resource  the Resource to get the qname for
	 * @return the qname or null
	 */
	public static String qnameFor(Resource resource) {
		Graph graph = resource.getModel().getGraph();
		if(graph instanceof MultiUnion) {
			String uri = resource.getURI();
	        int split = Util.splitNamespace(uri);
	        String local = uri.substring(split);
	        if (local.length() == 0) {
	        	return null;
	        }
	        String ns = uri.substring(0, split);
	        
	        MultiUnion mu = (MultiUnion) graph;
			Graph baseGraph = mu.getBaseGraph();
			if(baseGraph != null) {
				String prefix = baseGraph.getPrefixMapping().getNsURIPrefix(ns);
				if(prefix != null) {
					return prefix + ":" + local;
				}
			}
            List<Graph> graphs = mu.getSubGraphs();
            for (int i = 0; i < graphs.size(); i++) {
            	Graph subGraph = graphs.get(i);
            	String prefix = subGraph.getPrefixMapping().getNsURIPrefix(ns);
            	if(prefix != null && prefix.length() > 0) {
                	return prefix + ":" + local;
                }
            }
            return null;
		}
		else {
			return resource.getModel().qnameFor(resource.getURI());
		}
	}


	public void setIndentation(int value) {
		this.indentation = value;
	}
	
	
	public void setIndentationString(String value) {
		this.indentationString = value;
	}
	
	
	public void setNested(boolean value) {
		this.nested = value;
	}


	public void setPrintPrefixes(boolean value) {
		this.printPrefixes = value;
	}


	public void setUseExtraPrefixes(boolean value) {
		this.useExtraPrefixes = value;
	}
	
	
	public void setUsePrefixes(boolean value) {
		this.usePrefixes = value;
	}
}
