/*******************************************************************************
 * Copyright (c) 2009 TopQuadrant, Inc.
 * All rights reserved. 
 *******************************************************************************/
package org.topbraid.spin.system;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.topbraid.spin.arq.SPINARQFunction;
import org.topbraid.spin.arq.SPINARQPFunction;
import org.topbraid.spin.model.Function;
import org.topbraid.spin.model.SPINFactory;
import org.topbraid.spin.model.Template;
import org.topbraid.spin.util.JenaUtil;
import org.topbraid.spin.vocabulary.SPIN;
import org.topbraid.spin.vocabulary.SPL;

import com.hp.hpl.jena.graph.Graph;
import com.hp.hpl.jena.graph.compose.MultiUnion;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.sparql.function.FunctionRegistry;
import com.hp.hpl.jena.sparql.pfunction.PropertyFunctionRegistry;
import com.hp.hpl.jena.vocabulary.RDF;


/**
 * A singleton that keeps track of all registered SPIN functions
 * and templates.  For example, in TopBraid this is populated by
 * walking all .spin. files in the Eclipse workspace.  Other
 * implementations may need to register their modules "manually".
 * 
 * @author Holger Knublauch
 */
public class SPINModuleRegistry {
	
	/**
	 * Remembers all function definitions (in their original Model) so that they
	 * can be retrieved later.
	 */
	private Map<String, Function> functions = new HashMap<String, Function>();
	
	/**
	 * Remembers all template definitions (in their original Model) so that they
	 * can be retrieved later.
	 */
	private Map<String, Template> templates = new HashMap<String, Template>();

	
	private static SPINModuleRegistry singleton = new SPINModuleRegistry();
	
	
	/**
	 * Gets the singleton instance of this class.
	 * @return the singleton
	 */
	public static SPINModuleRegistry get() {
		return singleton;
	}
	
	
	/**
	 * Sets the SPINModuleRegistry to another value.
	 * @param value  the new value (not null)
	 */
	public static void set(SPINModuleRegistry value) {
		singleton = value;
	}
	
	
	/**
	 * Gets a registered Function with a given URI.
	 * @param uri  the URI of the Function to get
	 * @param model  an (optional) Model that should also be used to look up
	 *               locally defined functions (currently not used)
	 * @return the Function or null if none was found
	 */
	public Function getFunction(String uri, Model model) {
		Function function = functions.get(uri);
		if(function != null) {
			return function;
		}
		if(model != null) {
			function = model.getResource(uri).as(Function.class);
			if(JenaUtil.hasIndirectType(function, (Resource)SPIN.Function.inModel(model))) {
				return function;
			}
		}
		return null;
	}
	
	
	/**
	 * Gets a Template with a given URI in its defining Model.
	 * @param uri  the URI of the Template to look up
	 * @param model  an (optional) Model that should also be used for look up
	 * @return a Template or null
	 */
	public Template getTemplate(String uri, Model model) {
		if(model != null) {
			Resource r = model.getResource(uri);
			if(JenaUtil.hasIndirectType(r, (Resource)SPIN.Template.inModel(model))) {
				return r.as(Template.class);
			}
		}
		return templates.get(uri);
	}
	
	
	/**
	 * Gets a Collection of all registered Templates.
	 * @return the Templates
	 */
	public Collection<Template> getTemplates() {
		return templates.values();
	}
	
	
	/**
	 * Initializes this registry with all system functions and templates
	 * from the SPL namespace.
	 */
	public void init() {
		Model splModel = SPL.getModel();
		Model spinModel = SPIN.getModel();
		MultiUnion multiUnion = new MultiUnion(new Graph[] {
			splModel.getGraph(),
			spinModel.getGraph()
		});
		multiUnion.setBaseGraph(splModel.getGraph());
		Model unionModel = ModelFactory.createModelForGraph(multiUnion);
		registerAll(unionModel);
	}
	
	
	/**
	 * Registers a Function with its URI to this registry.
	 * As an optional side effect, if the provided function has a spin:body,
	 * this method can also register an ARQ FunctionFactory at the current
	 * Jena FunctionRegistry, using <code>registerARQFunction()</code>.
	 * <b>Note that the Model attached to the function should be an OntModel
	 * that also imports the system namespaces spin.owl and sp.owl - otherwise
	 * the system may not be able to transform the SPIN RDF into the correct
	 * SPARQL string.</b>
	 * @param function  the Function (must be a URI resource)
	 * @param addARQFunction  true to also add an entry to the ARQ function registry
	 */
	public void register(Function function, boolean addARQFunction) {
		functions.put(function.getURI(), function);
		ExtraPrefixes.add(function);
		if(addARQFunction) {
			registerARQFunction(function);
			if(function.hasProperty(RDF.type, SPIN.MagicProperty)) {
				registerARQPFunction(function);
			}
		}
	}
	
	
	/**
	 * Registers a Template with its URI.
	 * <b>Note that the Model attached to the template should be an OntModel
	 * that also imports the system namespaces spin.owl and sp.owl - otherwise
	 * the system may not be able to transform the SPIN RDF into the correct
	 * SPARQL string.</b>
	 * @param template  the Template (must be a URI resource)
	 */
	public void register(Template template) {
		templates.put(template.getURI(), template);
	}
	

	/**
	 * Registers all functions and templates from a given Model.
	 * <b>Note that the Model should contain the triples from the
	 * system namespaces spin.owl and sp.owl - otherwise the system
	 * may not be able to transform the SPIN RDF into the correct
	 * SPARQL string.  In a typical use case, the Model would be
	 * an OntModel that also imports the SPIN system namespaces.</b>
	 * @param model  the Model to iterate over
	 */
	public void registerAll(Model model) {
		registerFunctions(model);
		registerTemplates(model);
	}


	/**
	 * If the provided Function has an executable body (spin:body), then
	 * register an ARQ function for it with the current FunctionRegistry. 
	 * @param function  the function to register
	 */
	protected void registerARQFunction(Function function) {
		if(function.hasProperty(SPIN.body)) {
			final SPINARQFunction arqFunction = new SPINARQFunction(function);
			FunctionRegistry.get().put(function.getURI(), arqFunction);
		}
	}


	/**
	 * If the provided Function has an executable body (spin:body), then
	 * register an ARQ function for it with the current FunctionRegistry. 
	 * @param function  the function to register
	 */
	protected void registerARQPFunction(Function function) {
		if(function.hasProperty(SPIN.body)) {
			SPINARQPFunction arqFunction = new SPINARQPFunction(function);
			PropertyFunctionRegistry.get().put(function.getURI(), arqFunction);
		}
	}
	
	
	/**
	 * Registers all functions defined in a given Model.
	 * This basically iterates over all instances of spin:Function and calls
	 * <code>register(function)</code> for each of them.
	 * @param model  the Model to add the functions of
	 */
	public void registerFunctions(Model model) {
		for(Resource resource : JenaUtil.getAllInstances((Resource)SPIN.Function.inModel(model))) {
			Function function = SPINFactory.asFunction(resource);
			register(function, true);
		}
	}


	/**
	 * Registers all templates defined in a given Model.
	 * This basically iterates over all instances of spin:Template and calls
	 * <code>register(template)</code> for each of them.
	 * @param model  the Model to add the templates of
	 */
	public void registerTemplates(Model model) {
		for(Resource resource : JenaUtil.getAllInstances((Resource)SPIN.Template.inModel(model))) {
			if(resource.isURIResource()) {
				Template template = resource.as(Template.class);
				register(template);
				ExtraPrefixes.add(template);
			}
		}
	}
}
